# This package sets up the communication functions.  It calls the 
# appropriate comm package (UDP RS232 JTAG) depending on how it is initially configured.
#currently only supports RS232

package provide BTS_COMM 0.4;

package require BTS_BASE;

# ***********************************************************************
#	Init function: Sets up variables
# ***********************************************************************

namespace eval ::bts::comm {
	namespace import ::bts::bts
	
	bts add_config_function comm [list ::bts::comm::config];
	
	
	#table defines the write and read functions for various communication methods
	#only modes listed here are currently supported
	#variable bts_comm_write;
	#set bts_comm_write(rs232) "::bts::comm::rs232_write";
	
	#variable bts_comm_read;
	#set bts_comm_read(rs232) "::bts::comm::rs232_read";
	
	#variable bts_comm_open;
	#set bts_comm_open(rs232) "::bts::comm::rs232_open";
	
	#variable bts_comm_close(rs232);
	#set bts_comm_close(rs232) "::bts::comm::rs232_close";
	
	#the following variable is used to track return functions
	#for bts writes where data is expected back
	#the idea is to assign a number to these writes tracked on both
	#the host and the nios system.
	#the the nios system will return this id when it sends back data
	#The hash table below will in turn point to the user function 
	#that expect the data back
	variable bts_comm_ret_func;
	set bts_comm_ret_func(0) "::bts::comm::bts_comm_print";
	
	#open_port pointer
	variable bts_comm_port;
	#set bts_comm_port() "";
	
	#close_port pointer
	#variable bts_comm_close;
	#set bts_comm_close(rs232) "";
	
	variable bts_comm_settings;
	set bts_comm_settings(default) [list rs232 com1];
	set bts_comm_settings(rs232_com1) [list rs232 com1];
	set bts_comm_settings(rs232_com2) [list rs232 com2];
	set bts_comm_settings(rs232_ttyS0) [list rs232 /dev/ttyS0];
	set bts_comm_settings(rs232_ttyS1) [list rs232 /dev/ttyS1];
	
	variable bts_comm;
	set bts_comm(handler_state) "disabled";
	set bts_comm(bts_port) "";
	set bts_comm(debug) 1;
	
}

#config procedure for comm
proc ::bts::comm::config {{my_instruction ""} {var2 ""} {var3 ""} {var4 ""} {var5 ""} {var6 ""}} {
	variable bts_comm;
	variable bts_comm_settings;
	variable bts_comm_ret_func;
	
	if {$my_instruction == ""} {
		set my_instruction "help";
	}
	
	switch -exact -- $my_instruction {
		help {
			::bts::bts_base_print std "Available BTS COMM instructions are:\n";
			::bts::bts_base_print std "set open close write read\n";
		}		
		open {
			#call the open function
			if {[regexp $var2 [array names bts_comm_settings]]} {
				set return_val [::bts::comm::bts_open_port $var2];
				return $return_val;
			} else {
				::bts::bts_base_print error "There are no settings for $var2.\n";
			}
		}
		open_bts_port {
			#call the open function
			if {$bts_comm(bts_port) != ""} {
				set return_val [::bts::comm::bts_open_port $bts_comm(bts_port)];
				return $return_val;
			} else {
				::bts::bts_base_print error "There are no settings for bts_port.\n";
			}
		}
		close_bts_port {
			#call the close function
			if {[catch {set bts_comm(bts_port) $bts_comm(bts_port)}]} {
				if {[regexp $bts_comm(bts_port) [array names bts_comm_settings]]} {
					::bts::bts comm disable_read_handler;
					::bts::comm::bts_close_port $bts_comm(bts_port)	
				} else {
					::bts::bts_base_print error "There are no settings for $var2.\n";
				}
			} else {
				::bts::bts_base_print error "BTS Port is not open.\n";
			}
		}
		is_port_open {
			if {[catch {set bts_comm(bts_port) $bts_comm(bts_port)}]} {
				if {[regexp $bts_comm(bts_port) [array names bts_comm_settings]]} {
					return 1;
				} else {
					return 0;
				}
			} else {
				return 0;
			}
		}
		close {
			#call the close function
			if {[regexp $var2 [array names bts_comm_settings]]} {
				::bts::comm::bts_close_port $var2;	
			} else {
				::bts::bts_base_print error "There are no settings for $var2.\n";
			}
		}
		status {
			::bts::bts_base_print std "Current COMM Status is:\n";
			foreach index [array names bts_comm] {
				::bts::bts_base_print std "$index : $bts_comm($index)\n";
			}
			::bts::bts_base_print std "Current Port Settings are:\n";
			foreach index [array names bts_comm_settings] {
				::bts::bts_base_print std "$index : ";
				foreach element $bts_comm_settings($index) {
					::bts::bts_base_print std "$element ";
				}
				::bts::bts_base_print std "\n";
			}
		}
		set {
			if {[regexp $var2 [array names bts_comm]]} {
				
				set return_val [set bts_comm($var2) $var3];
			} else {
				::bts::bts_base_print error "$var2 is an invalid setting\n";
			}
		}
		add_com_port {
			set bts_comm_settings($var2) [list $var3 $var4 $var5 $var6];
		}
		write_port {
			#call the write function pointed to by 
			#the current comm type
			if {$var5 !=""} {
				#set $var5 "none";
			}
				set return_val [::bts::comm::bts_write_port $var2 $var3 $var4 $var5];
				return $return_val;
		}
		function_id_write_port {
			#
			#call the write function pointed to by 
			#the current comm type
			if {$var5 !=""} {
				#set $var5 "none";
			}
				set return_val [::bts::comm::bts_command_id_write_port $var2 $var3 $var4 $var5];
				return $return_val;
		}
		read_port {
			#call the read function pointed to by 
			#the current comm type
			if {$bts_comm(handler_state) == "disabled"} {
				::bts::comm::bts_read_port $var2;
				
			} else {
				::bts::bts_base_print error "Handler is enabled:  Disable Handler before doing a manual read!\n";
				return 1;
			}
		}
		enable_read_handler {
			if {$bts_comm(handler_state) == "disabled"} {

					set bts_comm(handler_state) "enabled";
					set return_val [::bts::comm::enable_read_handler];
					return $return_val;
				
			} else {
				::bts::bts_base_print error "Handler is already enabled\n";
			}
		}
		disable_read_handler {
			if {[regexp $var2 [array names bts_comm_port]]} {
				set bts_comm(handler_state) "disabled";
				set return_val [::bts::comm::disable_read_handler];
				return $return_val;
			} else {
				::bts::bts_base_print error "$var2 is an invalid port name.";
				return 1;
			}
		}
		set_cont_function_id {
			#expects -- command return_function_pointer
			#returns command_id for return function pointer
			
			#command id's 0 - 19 are saved for continuous return function assignments
			#command id 0 is reserved for error printing.
			set command_id 0;
			set return_function $var2
			#this loop looks for an open ret_func number
			set match 1;
			while {$match} {
				set command_id [expr $command_id + 1];
				set sub_match 0;
				foreach element [lsort -integer [array names bts_comm_ret_func]] {
					if {$element == $command_id} {
						set sub_match 1;
					}
				}
				if {$sub_match == 0} {
					set match 0;
				}
			}
			#return_Function better be correct or there will be
			#problems
			if {$command_id <20 } {
				#if it is in the reserved range....
				set bts_comm_ret_func($command_id) [list $return_function];
				return $command_id;
			} else {
				::bts::bts_base_print error "No valid continuous command ID found!!\n" 
				return 1; #there is an error
			}
			
		}
		unset_cont_function_id {
			#expects -- command command_id value
			
			#turn off a continuous ret functions
			#except for 0 -- reserved for error printing
			if {[regexp $var2 [array names bts_comm_ret_func]]} {
				if {$var2 != 0} {
					unset bts_comm_ret_func($var2);
				} else {
					::bts::bts_base_print error "Can't unset command_id 0\n";\
					return 1; #there is an error
				}
			} else {
				::bts::bts_base_print error "$var2 is not a valid command_id!\n";
				return 1; #there is an error
			}
		}
		default {
			::bts::bts_base_print error "$my_instruction is an invalid instruction\n";
			return 1; #there is an error
		}
	}
	
}

#rs232 function to open port
proc ::bts::comm::bts_open_port {port_name} {
	variable bts_comm_port;
	variable bts_comm_settings;
	set port_type [lindex $bts_comm_settings($port_name) 0];
	
	switch -exact -- $port_type {
		rs232 {
			#port_settings is list TYPE PORT
			set port [lindex $bts_comm_settings($port_name) 1];
			#set bts_comm_port($port_name) [open $port w+];
			if {[catch {open $port w+} bts_comm_port($port_name)]} {
				::bts::bts_base_print error "Can't open Port $port.\n";
			} else {
				#fconfigure $bts_comm_port($port_name) -mode 115200,n,8,1 -handshake xonxoff -buffering none -blocking 0;
				fconfigure $bts_comm_port($port_name) -mode 115200,n,8,1 -handshake xonxoff -buffering none -blocking 0;
				flush $bts_comm_port($port_name);
				return $bts_comm_port($port_name);
			}
		}
		default {
			::bts::bts_base_print error "Invalid communications type!\n";
			::bts::bts_base_print error "Valid types are: rs232\n"
			return 0;
		}	
	}

}

#function to close port
proc ::bts::comm::bts_close_port {port_name} {
	variable bts_comm_port;
	
	close $bts_comm_port($port_name);
	unset bts_comm_port($port_name);
}

#write function that adds command id
proc ::bts::comm::bts_write_port {command action_count send_data return_function} {
	variable bts_comm_port;
	variable bts_comm;
	variable bts_comm_ret_func;
	
	if {$return_function == "" } {
		set return_function "none";
	#	puts "none\n"
	}
	
	
	#if no return function is specified, point it to the default function
	#which will print only if debug is set
	if {$return_function == "none"} {	
		set command_id 0;
	} else {
		#command id's 0 - 19 are saved for continuous return function assignments
		set command_id 20;
		#this loop looks for an open ret_func number
		set match 1;
		while {$match} {
			set command_id [expr $command_id + 1];
			set sub_match 0;
			foreach element [lsort -integer [array names bts_comm_ret_func]] {
				if {$element == $command_id} {
					set sub_match 1;
				}
			}
			if {$sub_match == 0} {
				set match 0;
			}
		}
		#return_Function better be correct or there will be
		#problems
		set bts_comm_ret_func($command_id) [list $return_function];
	}
	
	#call the actual write function
	return [::bts::comm::bts_command_id_write_port $command_id $command $action_count $send_data];
	
}

#proc that octually does the write
proc ::bts::comm::bts_command_id_write_port {command_id command action_count send_data} {
	variable bts_comm_port;
	variable bts_comm;
	variable bts_comm_ret_func;
		
	#check if the command id exists, and error if it does not.  else send packet
	if {[regexp $command_id [array names bts_comm_ret_func]]} {
	
		set packet $command;
		lappend packet $action_count $command_id 0;
		
		#check if the port exists or is open
		#wierd, but it works
		if {[catch {set bts_comm_port($bts_comm(bts_port)) $bts_comm_port($bts_comm(bts_port))}]} {
			::bts::bts_base_print error "Port does not appear to be open!\n";
			return 1; #there is an error
		} else {
			
			#and finally send the packet
			#puts -nonewline $bts_comm_port($bts_comm(bts_port)) "S";
			puts -nonewline $bts_comm_port($bts_comm(bts_port)) "P";
			puts -nonewline $bts_comm_port($bts_comm(bts_port)) [join $packet ";"];
			foreach element $send_data {
				puts -nonewline $bts_comm_port($bts_comm(bts_port)) ";$element";
				#puts $element
			}
			puts $bts_comm_port($bts_comm(bts_port)) "N\n";
			return 0;
		}
		
	} else {
		::bts::bts_base_print error "Invalid Command_id Value!\n";
		return 1; #there is an error
		
	}
	
}


#read function -- not really to be used
proc ::bts::comm::bts_read_port {port_name {handler_flag 0}} {
	variable bts_comm;
	variable bts_comm_port;	
	variable bts_comm_ret_func;
	
	#this function is called everytime data arrives
	#i then decides where to send it
	
	#if we were called by fileevent
	if {$handler_flag} {
		set returned_data [gets $bts_comm_port($port_name)];
		
		#puts $returned_data;
		
		set capt_ret_data "";
		regexp {(\[[^\[]*?\])} $returned_data capt_ret_data;
		
		# if {$capt_ret_data != $returned_data} {
			# puts $returned_data;
		# }
		
		set data_list [split $capt_ret_data ";"];
		
		set match 0;
		set command_id [lindex $data_list 1];
		
		foreach index [array names bts_comm_ret_func] {
			if {$index == $command_id} {
				set match 1;
			}
		}
		#puts $match
		if {$match} {
			#return the data to the indicated return function
			set list_size [llength $data_list];
			set return_data_list [lrange $data_list 2 [expr $list_size - 2]];
			#puts $return_data_list;
			
			#unset the return function if it returns a non zero value
			if {[$bts_comm_ret_func($command_id) $return_data_list] > 19} {
				unset bts_comm_ret_func($command_id);
			}
		} else {
			#::bts::bts_base_print error $returned_data;
			$bts_comm_ret_func(0) $returned_data;
		}
		
	} else {
		set available_size [lindex [fconfigure $bts_comm_port($port_name) -queue] 0];
		if {$available_size} {
			set returned_data [read $bts_comm_port($port_name) $available_size];
		} else {
			set returned_data "";
		}				
	}
	
}

#enables event handler.  checks if a channel is readable, the listed
#function is called
proc ::bts::comm::enable_read_handler {} {
	variable bts_comm;
	variable bts_comm_port;

	fileevent $bts_comm_port($bts_comm(bts_port)) readable [list ::bts::comm::bts_read_port $bts_comm(bts_port) 1];
		
}

proc ::bts::comm::disable_read_handler {} {
	variable bts_comm;
	variable bts_comm_port;

	fileevent $bts_comm_port($bts_comm(bts_port))  readable "";
		
}


proc ::bts::comm::bts_comm_print {text} {
	variable bts_comm;
	# if {$bts_comm(debug)} {
		# ::bts::bts_base_print std "Received: \n"
		# foreach index $text {
			# ::bts::bts_base_print std "$index \n";
		# }
		# ::bts::bts_base_print std "\n";
	# }
	
	::bts::bts_base_print std "$text\n";
}
